/* read the README.virtual for more info on using the
*  virtual patch for gnu-pop3d */

#include "gnu-pop3d.h"

#ifdef USE_VIRTUAL

#ifndef VIRTUAL_PASSWORDS_PATH
#define VIRTUAL_PASSWORDS_PATH  "/etc/virtual"
#endif

#ifndef VIRTUAL_PASSWORD_FNAME
#define VIRTUAL_PASSWORD_FNAME  "passwd"
#endif

struct passwd *
getvirtualpwnam (const char *username, const char *domainname)
{
  FILE *passwdfd;
  char buffer[256 * sizeof (char)];
  struct passwd *pwdentry = NULL;
  int found = 0;

  snprintf (buffer, 256, "%s/%s/%s", VIRTUAL_PASSWORDS_PATH, domainname, VIRTUAL_PASSWORD_FNAME);
  if (debug & 0x40)
	syslog (LOG_INFO, "Checking %s for %s (%s)", buffer, username, domainname);

  if (((pwdentry =
		(struct passwd *) calloc (1, sizeof (struct passwd))) == NULL)
	  || ((pwdentry->pw_name = (char *) calloc (256, sizeof (char))) == NULL)
	  || ((pwdentry->pw_passwd = (char *) calloc (256, sizeof (char))) ==
		  NULL)
	  || ((pwdentry->pw_dir = (char *) calloc (256, sizeof (char))) == NULL)
	  || ((pwdentry->pw_gecos = (char *) calloc (256, sizeof (char))) == NULL)
	  || ((pwdentry->pw_shell = (char *) calloc (256, sizeof (char))) == NULL))
	{
	  syslog (LOG_ERR, "Couldn't allocate memory for password check: %s (%s)", username, domainname);
	  pop3_freepw (pwdentry);
	  return (NULL);
	}

  if ((passwdfd = fopen (buffer, "r")) == (FILE *) NULL)
	{
	  syslog (LOG_ERR, "Couldn't open password file: %s", buffer);
	  pop3_freepw (pwdentry);
	  return (NULL);
	}
  while (fgets (buffer, sizeof (buffer), passwdfd) != (char *) NULL)
	{
	  if (debug & 0x80)
		syslog (LOG_ERR, "Line read: %s", buffer);

	  /* skip comments, blank lines */
	  if (*buffer == '#' || *buffer == '\n' || *buffer == '\r' || *buffer == '\0')
		{
		  continue;
		}

	  /* kristian:password18:1511:100:Kristian V. Thomsen:.: */
	  /* kristian:$5$S8$BbsHepgqbYsO43UhxMLZmWEAg1Kv.KxLgs.1IIJza97:1511:100:Kristian V. Thomsen:.: */

	  pwdentry->pw_shell[0] = '\0';
	  if (sscanf (buffer, "%[^:]:%[^:]:%u:%u:%[^:]:%[^:]:%s",
				  pwdentry->pw_name, pwdentry->pw_passwd, &pwdentry->pw_uid,
				  &pwdentry->pw_gid, pwdentry->pw_gecos, pwdentry->pw_dir, pwdentry->pw_shell) < 6)
		{
		  syslog (LOG_ERR, "%s: Line format in passwd file not new format %d", buffer, uid);

		  strcpy (pwdentry->pw_shell, "encrypted");	/* flag for old format password file */
		  pwdentry->pw_uid = uid;
		  pwdentry->pw_gid = gid;
		  pwdentry->pw_gecos[0] = '\0';
		  pwdentry->pw_dir[0] = '\0';

		  if (sscanf (buffer, "%[^:]:%s", pwdentry->pw_name, pwdentry->pw_passwd) != 2)
			{
			  syslog (LOG_ERR, "%s: Line format in passwd file incorrect \n", buffer);
			  continue;
			}
		}
	  if ((strlen (pwdentry->pw_name) == 0) || (strlen (pwdentry->pw_passwd) == 0) || (pwdentry->pw_uid <= 0))
		{
		  syslog (LOG_ERR, "%s: Trouble parsing passwd file\n", buffer);
		  continue;
		}
	  if (strcmp (pwdentry->pw_name, username) == 0)
		{
		  found = 1;
		  break;
		}
	}
  (void) fclose (passwdfd);
/*  fprintf (stderr, "uid %d gid %d\n", pwdentry->pw_uid, pwdentry->pw_gid); */

  if (found)
	return pwdentry;
  else
	return (NULL);
}
#endif /* USE_VIRTUAL */

/* Split username into username and virtualdomain and possible useroptions
name[{;!@:}domain][--options] */
char *
pop3_splitname (char *arg, char *useroptions)
{
  char *temp_domain, *virtualdomain;
  const char domain_delimiters[] = ";!:@%";
  int i;
#ifdef IP_BASED_VIRTUAL
  struct sockaddr_storage sockname;
  socklen_t namelen = sizeof (sockname);
  char sockipaddr[(INET6_ADDRSTRLEN + 1) * sizeof (char)];
  char host[MAXHOSTNAMELEN + 1];
  char service[32 * sizeof (char)];

#endif

  virtualdomain = NULL;
  useroptions = NULL;

  /* start by getting any options */
  if ((useroptions = strstr (arg, "--")) != NULL)
	{
	  *useroptions = '\0';
	  useroptions++;
	  useroptions++;
	  useroptions = strdup (useroptions);

/* turn message disposition notification (read receipts) on/off */
	  if (strstr (useroptions, "r+") != NULL)
		mdn = 1;
	  if (strstr (useroptions, "r-") != NULL)
		mdn = 0;

	  if (strlen (useroptions) == 0)
		useroptions = NULL;
	  if (debug & 0x01)
		syslog (LOG_INFO, "Useroptions: %s (%s)", useroptions, arg);
	}

  /* Some mail clients pre-parse out the username before sending */
  /* so allow colon or at-sign to separate username and domain name */

  for (i = strlen (domain_delimiters) - 1; i >= 0; i--)
	{
	  temp_domain = strchr (arg + 1, domain_delimiters[i]);
	  if (temp_domain && (strlen (temp_domain) > 1))
		{
		  temp_domain[0] = '\0';	/* arg becomes username without hostname */
		  temp_domain++;
		  if (virtualdomain)
			{
			  free (virtualdomain);
			}
		  virtualdomain = strdup (temp_domain);
		  break;
		}
	}

/* placing domain name in USER command overrides IP_BASED_VIRTUAL */

#ifdef IP_BASED_VIRTUAL
  if (virtualdomain == NULL)
	{
	  if (getsockname (ifile, (struct sockaddr *) &sockname, &namelen) != 0)
		{
		  syslog (LOG_ERR, "getsockname: %s", strerror (errno));
		  pop3_abquit (ERR_IP_BASED);
		}
	  sockipaddr[0] = '\0';
	  if (sockname.ss_family == AF_INET)
		inet_ntop (sockname.ss_family, &((struct sockaddr_in *) &sockname)->sin_addr.s_addr, sockipaddr,
				   sizeof (sockipaddr));
#ifdef USE_IPV6
	  else if (sockname.ss_family == AF_INET6)
		inet_ntop (sockname.ss_family, &((struct sockaddr_in6 *) &sockname)->sin6_addr, sockipaddr,
				   sizeof (sockipaddr));
#endif
	  if (sockipaddr[0] == '\0')
		{
		  syslog (LOG_ERR, "inet_ntop failed: %s", strerror (errno));
		  pop3_abquit (ERR_IP_BASED);
		}
	  if (debug & 0x40)
		syslog (LOG_INFO, "Connection to %s", sockipaddr);

	  if (!getnameinfo ((struct sockaddr *) &sockname, namelen, host, sizeof (host), service, sizeof (service), NI_IDN))
		{
		  if (!virtualdomain)
			{
			  virtualdomain = malloc (MAXHOSTNAMELEN + 1);
			  if (virtualdomain == NULL)
				pop3_abquit (ERR_NO_MEM);
			}
		  strcpy (virtualdomain, host);
		  if (debug & 0x40)
			syslog (LOG_INFO, "hostname: %s", virtualdomain);
		}
	  else
		{
		  pop3_abquit (ERR_IP_BASED);
		  syslog (LOG_ERR, "getnameinfo failure %d for %s\n", h_errno, temp_domain);
		}
	}
#endif /* IP_BASED_VIRTUAL */

  return virtualdomain;
}
